#include "rasteriser/rasteriser.h"
#include "aliases.h"
#include "list/typed_list.h"
#include "math/math_utils.h"
#include "mem_arena.h"
#include "vector/vec.h"
#include "window/window.h"
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>

internal void render_instance(window_t *wnd, const rasteriser_scene_t *scene,
                              Arena *arena, mat4x4f_t camera_matrix,
                              mat3x4f_t projection_matrix,
                              const instance_t *instance, render_type_t type);
internal void clip_instance(Arena *arena, model_t *model,
                            const rasteriser_scene_t *scene);
internal void clip_instance_against_plane(Arena *arena, model_t *model,
                                          const bounding_sphere_t *sphere,
                                          const clipping_plane_t *plane);
internal void clip_triangles_against_plane(Arena *arena, model_t *model,
                                           const clipping_plane_t *plane);
internal void clip_triangle_against_plane(Arena *arena, model_t *model,
                                          i64 triangle_index,
                                          const clipping_plane_t *plane);
internal list_float *interpolate(Arena *arena, i32 i0, f32 d0, i32 i1, f32 d1);
internal inline void order_triangle_points(triangle_t *triangle);
internal inline bounding_sphere_t
get_bounding_sphere(const list_vertex_t *vertices);
internal inline vertex_t segment_intersection(const vertex_t *a,
                                              const vertex_t *b,
                                              const clipping_plane_t *plane);
internal inline void copy_triangles(Arena *arena, list_scene_triangle_t *dst,
                                    const list_scene_triangle_t *src);
internal inline f32 signed_distance(const vec3f_t *vertex,
                                    const clipping_plane_t *plane);

void render_scene(window_t *wnd, Arena *arena, const rasteriser_scene_t *scene,
                  render_type_t type) {
  mat4x4f_t cam_inverse_rotation =
      inverse_camera_rotation_matrix(scene->camera);
  mat4x4f_t cam_inverse_translation =
      inverse_camera_translation_matrix(scene->camera);
  mat4x4f_t cam_matrix =
      mul_mat4x4f(cam_inverse_rotation, cam_inverse_translation);

  for (u64 i = 0; i < scene->instance_count; ++i) {
    render_instance(wnd, scene, arena, cam_matrix, *(scene->proj_matrix),
                    &(scene->instances[i]), type);
  }
}

internal void render_instance(window_t *wnd, const rasteriser_scene_t *scene,
                              Arena *arena, mat4x4f_t camera_matrix,
                              mat3x4f_t projection_matrix,
                              const instance_t *instance, render_type_t type) {
  list_vertex_t *transformed = list_create_with_capacity(
      vertex_t, arena,
      (instance->model->vertices->count + 10) * sizeof(vertex_t));
  list_scene_triangle_t *triangles = list_create_with_capacity(
      scene_triangle_t, arena,
      (instance->model->triangles->count + 10) * sizeof(scene_triangle_t));
  if (!transformed || !triangles) {
    return;
  }

  copy_triangles(arena, triangles, instance->model->triangles);

  mat4x4f_t t_mat = get_translation_matrix(instance->transform.position);
  mat4x4f_t r_mat = get_rotation_matrix(instance->transform.rotation);
  mat4x4f_t s_mat = get_scaling_matrix((vec3f_t){instance->transform.scale,
                                                 instance->transform.scale,
                                                 instance->transform.scale});
  mat4x4f_t tr_mat = mul_mat4x4f(t_mat, r_mat);
  mat4x4f_t trs_mat = mul_mat4x4f(tr_mat, s_mat);

  mat4x4f_t xf_cam_mat = mul_mat4x4f(camera_matrix, trs_mat);

  vertex_t vertex = {0};
  vec4f_t xf_vertex = {0};
  for (u64 i = 0; i < instance->model->vertices->count; ++i) {
    vertex = list_get(instance->model->vertices, i);

    xf_vertex = mul_mat4x4f_by_vec4f(
        xf_cam_mat, (vec4f_t){vertex.position.x, vertex.position.y,
                              vertex.position.z, 1.0f});
    xf_vertex = vec_div_num(vec4f_t, xf_vertex, xf_vertex.w);

    vertex.position = (vec3f_t){xf_vertex.x, xf_vertex.y, xf_vertex.z};

    list_append(vertex_t, arena, transformed, vertex);
  }

  model_t model = {transformed, triangles};
  clip_instance(arena, &model, scene);
  if (is_null_model(&model)) {
    return;
  }

  list_vec2i_t *projected = list_create_with_capacity(
      vec2i_t, arena, instance->model->vertices->count * sizeof(vec2i_t));
  if (!projected) {
    return;
  }

  vec2i_t point = {0};
  for (u64 i = 0; i < transformed->count; ++i) {
    vertex = list_get(transformed, i);
    vertex.position = mul_mat3x4f_by_vec4f(
        projection_matrix, (vec4f_t){vertex.position.x, vertex.position.y,
                                     vertex.position.z, 1.0f});
    point = (vec2i_t){(i32)(vertex.position.x / vertex.position.z),
                      (i32)(vertex.position.y / vertex.position.z)};
    list_append(vec2i_t, arena, projected, point);
  }

  vertex_t v0, v1, v2;
  for (u64 i = 0; i < triangles->count; ++i) {
    scene_triangle_t triangle = list_get(triangles, i);
    v0 = list_get(transformed, triangle.idx0);
    v1 = list_get(transformed, triangle.idx1);
    v2 = list_get(transformed, triangle.idx2);

    triangle_t tri = {
        .p0 = list_get(projected, triangle.idx0),
        .p1 = list_get(projected, triangle.idx1),
        .p2 = list_get(projected, triangle.idx2),
        .h0 = v0.h,
        .h1 = v1.h,
        .h2 = v2.h,
        .z0 = 1.0f / v0.position.z,
        .z1 = 1.0f / v1.position.z,
        .z2 = 1.0f / v2.position.z,
        .colour = triangle.colour,
    };

    if ((tri.p0.x == tri.p1.x && tri.p0.x == tri.p2.x) &&
        (tri.p0.y == tri.p1.y && tri.p0.y == tri.p2.y)) {
      continue;
    }

    switch (type) {
    case RASTERISER_RENDER_WIREFRAME:
      draw_wireframe_triangle(wnd, arena, tri);
      break;
    case RASTERISER_RENDER_FILLED:
      draw_filled_triangle(wnd, arena, tri);
      break;
    case RASTERISER_RENDER_SHADED:
      draw_shaded_triangle(wnd, arena, tri);
      break;
    }
  }
}

void draw_wireframe_triangle(window_t *wnd, Arena *arena, triangle_t triangle) {
  order_triangle_points(&triangle);

  draw_line(wnd, arena, (line_t){triangle.p0, triangle.p1}, triangle.colour);
  draw_line(wnd, arena, (line_t){triangle.p1, triangle.p2}, triangle.colour);
  draw_line(wnd, arena, (line_t){triangle.p2, triangle.p0}, triangle.colour);
}

void draw_filled_triangle(window_t *wnd, Arena *arena, triangle_t triangle) {
  order_triangle_points(&triangle);

  i32 x0 = triangle.p0.x;
  i32 y0 = triangle.p0.y;
  i32 x1 = triangle.p1.x;
  i32 y1 = triangle.p1.y;
  i32 x2 = triangle.p2.x;
  i32 y2 = triangle.p2.y;

  list_float *x01 = interpolate(arena, y0, x0, y1, x1);
  list_float *x12 = interpolate(arena, y1, x1, y2, x2);
  list_float *x02 = interpolate(arena, y0, x0, y2, x2);
  list_float *x012 = NULL;

  list_pop(x01); // Last element of x01 is a duplicate of first element in x12
  list_merge(f32, arena, x012, x01, x12);

  f32 z0 = triangle.z0;
  f32 z1 = triangle.z1;
  f32 z2 = triangle.z2;
  list_float *z01 = interpolate(arena, y0, z0, y1, z1);
  list_float *z12 = interpolate(arena, y1, z1, y2, z2);
  list_float *z02 = interpolate(arena, y0, z0, y2, z2);
  list_float *z012 = NULL;

  list_pop(z01); // Last element of z01 is a duplicate of first element in z12
  list_merge(f32, arena, z012, z01, z12);

  list_float *x_left;
  list_float *x_right;
  list_float *z_left;
  list_float *z_right;
  u64 middle = (u64)(floorf((f32)(x02->count) / 2.0f));
  if (list_get(x02, middle) < list_get(x012, middle)) {
    x_left = x02;
    z_left = z02;
    x_right = x012;
    z_right = z012;
  } else {
    x_left = x012;
    z_left = z012;
    x_right = x02;
    z_right = z02;
  }

  list_float *z_segment = NULL;
  i32 index = -1;
  i64 xl = -1;
  i64 xr = -1;
  f32 current_z = INFINITY;
  f32 new_z = INFINITY;

  for (i64 y = y0; y <= y2; ++y) {
    index = y - y0;
    xl = (i64)list_get(x_left, index);
    xr = (i64)list_get(x_right, index);

    for (i64 x = xl; x < xr; ++x) {
      z_segment = interpolate(arena, xl, list_get(z_left, index), xr,
                              list_get(z_right, index));

      current_z = get_z_pixel(wnd, x, y);
      new_z = list_get(z_segment, x - xl);

      if (new_z > current_z) {
        set_z_pixel(wnd, x, y, new_z);
        set_pixel(wnd, x, y, triangle.colour);
      }
    }
  }
}

void draw_shaded_triangle(window_t *wnd, Arena *arena, triangle_t triangle) {
  order_triangle_points(&triangle);

  i32 x0 = triangle.p0.x;
  i32 y0 = triangle.p0.y;
  i32 x1 = triangle.p1.x;
  i32 y1 = triangle.p1.y;
  i32 x2 = triangle.p2.x;
  i32 y2 = triangle.p2.y;

  list_float *x01 = interpolate(arena, y0, x0, y1, x1);
  list_float *x12 = interpolate(arena, y1, x1, y2, x2);
  list_float *x02 = interpolate(arena, y0, x0, y2, x2);
  list_float *x012 = NULL;

  list_pop(x01); // Last element of x01 is a duplicate of first element in x12
  list_merge(f32, arena, x012, x01, x12);

  f32 h0 = triangle.h0;
  f32 h1 = triangle.h1;
  f32 h2 = triangle.h2;
  list_float *h01 = interpolate(arena, y0, h0, y1, h1);
  list_float *h12 = interpolate(arena, y1, h1, y2, h2);
  list_float *h02 = interpolate(arena, y0, h0, y2, h2);
  list_float *h012 = NULL;

  list_pop(h01); // Last element of h01 is a duplicate of first element in h12
  list_merge(f32, arena, h012, h01, h12);

  f32 z0 = triangle.z0;
  f32 z1 = triangle.z1;
  f32 z2 = triangle.z2;
  list_float *z01 = interpolate(arena, y0, z0, y1, z1);
  list_float *z12 = interpolate(arena, y1, z1, y2, z2);
  list_float *z02 = interpolate(arena, y0, z0, y2, z2);
  list_float *z012 = NULL;

  list_pop(z01); // Last element of z01 is a duplicate of first element in z12
  list_merge(f32, arena, z012, z01, z12);

  list_float *x_left;
  list_float *x_right;
  list_float *h_left;
  list_float *h_right;
  list_float *z_left;
  list_float *z_right;
  u64 middle = (u64)(floorf((f32)(x02->count) / 2.0f));
  if (list_get(x02, middle) < list_get(x012, middle)) {
    x_left = x02;
    h_left = h02;
    z_left = z02;
    x_right = x012;
    h_right = h012;
    z_right = z012;
  } else {
    x_left = x012;
    h_left = h012;
    z_left = z012;
    x_right = x02;
    h_right = h02;
    z_right = z02;
  }

  list_float *h_segment = NULL;
  list_float *z_segment = NULL;
  i32 index = -1;
  i64 xl = -1;
  i64 xr = -1;
  f32 current_z = INFINITY;
  f32 new_z = INFINITY;
  colour_t shaded_colour = (colour_t){0};

  for (i64 y = y0; y <= y2; ++y) {
    index = y - y0;
    xl = (i64)list_get(x_left, index);
    xr = (i64)list_get(x_right, index);
    for (i64 x = xl; x < xr; ++x) {
      h_segment = interpolate(arena, xl, list_get(h_left, index), xr,
                              list_get(h_right, index));
      z_segment = interpolate(arena, xl, list_get(z_left, index), xr,
                              list_get(z_right, index));
      shaded_colour = colour_mul(triangle.colour, list_get(h_segment, x - xl));

      current_z = get_z_pixel(wnd, x, y);
      new_z = list_get(z_segment, x - xl);

      if (new_z > current_z) {
        set_z_pixel(wnd, x, y, new_z);
        set_pixel(wnd, x, y, shaded_colour);
      }
    }
  }
}

void draw_line(window_t *wnd, Arena *arena, line_t line, colour_t colour) {
  list_float *values = NULL;

  if (abs(line.p1.x - line.p0.x) > abs(line.p1.y - line.p0.y)) {
    if (line.p1.x < line.p0.x) {
      swap(vec2i_t, line.p0, line.p1);
    }

    i32 x0 = line.p0.x;
    i32 y0 = line.p0.y;
    i32 x1 = line.p1.x;
    i32 y1 = line.p1.y;

    values = interpolate(arena, x0, y0, x1, y1);
    if (!values) {
      return;
    }

    for (i32 x = x0; x <= x1; ++x) {
      set_pixel(wnd, x, (i32)(list_get(values, x - x0)), colour);
    }
  } else {
    if (line.p1.y < line.p0.y) {
      swap(vec2i_t, line.p0, line.p1);
    }

    i32 x0 = line.p0.x;
    i32 y0 = line.p0.y;
    i32 x1 = line.p1.x;
    i32 y1 = line.p1.y;

    values = interpolate(arena, y0, x0, y1, x1);
    if (!values) {
      return;
    }

    for (i32 y = y0; y <= y1; ++y) {
      set_pixel(wnd, (i32)(list_get(values, y - y0)), y, colour);
    }
  }
}

bool is_null_model(const model_t *model) {
  return model->vertices == NULL && model->triangles == NULL;
}

internal void clip_instance(Arena *arena, model_t *model,
                            const rasteriser_scene_t *scene) {
  bounding_sphere_t sphere = get_bounding_sphere(model->vertices);
  const clipping_plane_t *plane = NULL;

  for (u64 i = 0; i < CLIPPING_PLANE_COUNT; ++i) {
    plane = &(scene->planes[i]);
    clip_instance_against_plane(arena, model, &sphere, plane);
    if (is_null_model(model)) {
      return;
    }
  }
}

internal void clip_instance_against_plane(Arena *arena, model_t *model,
                                          const bounding_sphere_t *sphere,
                                          const clipping_plane_t *plane) {
  f32 distance = signed_distance(&(sphere->centre), plane);
  f32 neg_radius = -1.0f * sphere->radius;

  if (distance > sphere->radius) {
    return;
  } else if (distance < neg_radius) {
    *model = NULL_MODEL;
  } else {
    clip_triangles_against_plane(arena, model, plane);
  }
}

internal void clip_triangles_against_plane(Arena *arena, model_t *model,
                                           const clipping_plane_t *plane) {
  u64 count = model->triangles->count;
  for (u64 i = 0; i < count; ++i) {
    clip_triangle_against_plane(arena, model, i, plane);
  }
}

internal void clip_triangle_against_plane(Arena *arena, model_t *model,
                                          i64 triangle_index,
                                          const clipping_plane_t *plane) {
  scene_triangle_t *triangle = &(list_get(model->triangles, triangle_index));
  vertex_t *v0 = &(list_get(model->vertices, triangle->idx0));
  vertex_t *v1 = &(list_get(model->vertices, triangle->idx1));
  vertex_t *v2 = &(list_get(model->vertices, triangle->idx2));

  f32 d0 = signed_distance(&(v0->position), plane);
  f32 d1 = signed_distance(&(v1->position), plane);
  f32 d2 = signed_distance(&(v2->position), plane);

  if (d0 >= 0.0f && d1 >= 0.0f && d2 >= 0.0f) {
    return;
  } else if (d0 < 0.0f && d1 < 0.0f && d2 < 0.0f) {
    triangle->idx0 = INVALID_VERTEX;
    triangle->idx1 = INVALID_VERTEX;
    triangle->idx2 = INVALID_VERTEX;
  } else if ((d0 < 0.0f && d1 < 0.0f) || (d0 < 0.0f && d2 < 0.0f) ||
             (d1 < 0.0f && d2 < 0.0f)) {
    vertex_t *vp;
    vertex_t *vn0;
    vertex_t *vn1;
    i64 *in0;
    i64 *in1;
    if (d0 > 0.0f) {
      vp = v0;
      vn0 = v1;
      vn1 = v2;
      in0 = &(triangle->idx1);
      in1 = &(triangle->idx2);
    } else if (d1 > 0.0f) {
      vp = v1;
      vn0 = v0;
      vn1 = v2;
      in0 = &(triangle->idx0);
      in1 = &(triangle->idx2);
    } else {
      vp = v2;
      vn0 = v0;
      vn1 = v1;
      in0 = &(triangle->idx0);
      in1 = &(triangle->idx1);
    }

    vertex_t r0 = segment_intersection(vp, vn0, plane);
    vertex_t r1 = segment_intersection(vp, vn1, plane);
    list_append(vertex_t, arena, model->vertices, r0);
    list_append(vertex_t, arena, model->vertices, r1);

    *in0 = model->vertices->count - 2;
    *in1 = model->vertices->count - 1;
  } else {
    vertex_t *vn;
    vertex_t *vp0;
    vertex_t *vp1;
    i64 *in;
    i64 *ip;
    if (d0 < 0.0f) {
      vn = v0;
      vp0 = v1;
      vp1 = v2;
      in = &(triangle->idx0);
      ip = &(triangle->idx2);
    } else if (d1 < 0.0f) {
      vn = v1;
      vp0 = v0;
      vp1 = v2;
      in = &(triangle->idx1);
      ip = &(triangle->idx2);
    } else {
      vn = v2;
      vp0 = v0;
      vp1 = v1;
      in = &(triangle->idx2);
      ip = &(triangle->idx1);
    }

    vertex_t r0 = segment_intersection(vp0, vn, plane);
    vertex_t r1 = segment_intersection(vp1, vn, plane);
    list_append(vertex_t, arena, model->vertices, r0);
    list_append(vertex_t, arena, model->vertices, r1);

    *in = model->vertices->count - 2;

    scene_triangle_t tri = {
        .idx0 = *in,
        .idx1 = *ip,
        .idx2 = model->vertices->count - 1,
        .colour = triangle->colour,
    };

    list_append(scene_triangle_t, arena, model->triangles, tri);
  }
}

internal list_float *interpolate(Arena *arena, i32 i0, f32 d0, i32 i1, f32 d1) {
  list_float *values;
  if (i0 == i1) {
    values = list_create_with_capacity(f32, arena, 20);
    if (values) {
      list_append(f32, arena, values, d0);
    }

    return values;
  }

  values = list_create(f32, arena);
  if (!values) {
    return NULL;
  }

  f32 a = (d1 - d0) / ((f32)i1 - i0);
  f32 d = d0;

  for (i32 i = i0; i <= i1; ++i) {
    list_append(f32, arena, values, d);
    d += a;
  }

  return values;
}

internal inline void order_triangle_points(triangle_t *triangle) {
  if (triangle->p1.y < triangle->p0.y) {
    swap(vec2i_t, triangle->p0, triangle->p1);
    swap(f32, triangle->h0, triangle->h1);
  }
  if (triangle->p2.y < triangle->p0.y) {
    swap(vec2i_t, triangle->p0, triangle->p2);
    swap(f32, triangle->h0, triangle->h2);
  }
  if (triangle->p2.y < triangle->p1.y) {
    swap(vec2i_t, triangle->p1, triangle->p2);
    swap(f32, triangle->h1, triangle->h2);
  }
}

internal inline bounding_sphere_t
get_bounding_sphere(const list_vertex_t *vertices) {
  vec3f_t sum = {0};
  for (u64 i = 0; i < vertices->count; ++i) {
    sum = vec_add(vec3f_t, list_get(vertices, i).position, sum);
  }

  vec3f_t centre = vec_div_num(vec3f_t, sum, vertices->count);
  f32 radius = 0.0f;
  f32 tmp = 0.0f;
  vec3f_t dst = {0};
  for (u64 i = 0; i < vertices->count; ++i) {
    dst = vec_sub(vec3f_t, list_get(vertices, i).position, centre);
    tmp = vec_magnitude(vec3f_t, dst);
    radius = tmp > radius ? tmp : radius;
  }

  return (bounding_sphere_t){
      .centre = centre,
      .radius = radius,
  };
}

internal inline vertex_t segment_intersection(const vertex_t *a,
                                              const vertex_t *b,
                                              const clipping_plane_t *plane) {
  f32 neg_d = plane->distance * -1.0f;
  f32 na_dot = vec_dot(vec3f_t, plane->normal, a->position);
  vec3f_t ba = vec_sub(vec3f_t, b->position, a->position);
  f32 nba_dot = vec_dot(vec3f_t, plane->normal, ba);

  f32 t = (neg_d - na_dot) / nba_dot;

  return (vertex_t){
      .position = vec_add(vec3f_t, a->position, vec_mul_num(vec3f_t, ba, t)),
      .h = a->h + t * (b->h - a->h),
  };
}

internal inline void copy_triangles(Arena *arena, list_scene_triangle_t *dst,
                                    const list_scene_triangle_t *src) {
  for (u64 i = 0; i < src->count; ++i) {
    scene_triangle_t tri = list_get(src, i);
    list_append(scene_triangle_t, arena, dst, tri);
  }
}

internal inline f32 signed_distance(const vec3f_t *vertex,
                                    const clipping_plane_t *plane) {
  return (vertex->x * plane->normal.x) + (vertex->y * plane->normal.y) +
         (vertex->z * plane->normal.z) + plane->distance;
}
